/*
 * Acknowledgement:
 *     The methods used here for detecting UARTs and distinguishing UART
 * types were described in Jeff Duntemann's "Structured Programming"
 * column, in Dr. Dobb's Journal, July 1991.
 */

/*
 * UART type differences:
 *     8250 register updating and I-O may be too slow for fast CPUs.
 *     16450 register updating is OK, and it has a scratch register.
 *     16550 is like the 16450, and it almost has a FIFO.
 *     16550A has a working 16-byte buffer (FIFO).
 */

/*
 * Copyright (c) 1991 Thomas W. S. Smith.  All rights reserved.
 *
 * Written by Thomas W. S. Smith, September 26, 1991.
 *
 * Redistribution and use in source and binary forms are permitted provided
 * the above copyright notice and attribution and date of work and this
 * paragraph are duplicated in all such forms.  THIS SOFTARE IS PROVIDED
 * "AS IS" AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT
 * LIMITATION, THE IMPLIED WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE.
 */

#include <dos.h>
#include <stdio.h>
#include <stdlib.h>

enum uart_type {
	u_unknown = -1,
	u_none,
	u_8250,
	u_16450,
	u_16550,
	u_16550a
};

/*
 * UART register offsets
 */
#define RBR 0        /* R    receive buffer register */
#define THR 0        /*  W   transmit holding register */
#define IER 1        /* RW   interrupt enable register */
#define IIR 2        /* R    interrupt ID register */
#define FCR 2        /*  W   FIFO control register (16550 or later) */
#define LCR 3        /* RW   line control register */
#define MCR 4        /* RW   MODEM control register */
#define LSR 5        /* RW   line status register */
#define MSR 6        /* RW   MODEM status register */
#define SCR 7        /* RW   scratch register (16450 or later) */
#define DLL 0        /* RW   divisor latch LSB register */
#define DLM 1        /* RW   divisor latch MSB register */

/*
 * interrupt enable register
 */
#define IER_MS   0x08        /* MODEM status */
#define IER_RLS  0x04        /* Receiver Line Status */
#define IER_THRE 0x02        /* Transmit holding register empty */
#define IER_DR   0x01        /* Data ready */

/*
 * interrupt ID register
 */
#define IIR_RLS  0x06        /* Receiver Line Status */
#define IIR_RDA  0x04        /* Received data available */
#define IIR_THRE 0x02        /* Transmit holding register empty */
#define IIR_MS   0x00        /* MODEM status */
#define IIR_IP   0x01        /* Interrupt pending */

/*
 * FIFO control register
 */

/*
 * line control register
 */
#define LCR_DLAB     0x80        /* divisor latch access bit */
#define LCR_SBRK     0x40        /* send break */
#define LCR_DATA_5   0x00        /* data bits select */
#define LCR_DATA_6   0x01
#define LCR_DATA_7   0x02
#define LCR_DATA_8   0x03
#define LCR_PAR_NONE 0x00        /* parity select */
#define LCR_PAR_ODD  0x08
#define LCR_PAR_EVEN 0x18
#define LCR_PAR_ONE  0x28
#define LCR_PAR_ZERO 0x38
#define LCR_STOP_1   0x00        /* stop bits select */
#define LCR_STOP_2   0x04

/*
 * MODEM control register
 */
#define MCR_LOOP 0x10        /* loopback test mode */
#define MCR_OUT2 0x08
#define MCR_OUT1 0x04
#define MCR_RTS  0x02
#define MCR_DTR  0x01

/*
 * line status register
 */
#define LSR_THRE 0x40        /* transmitter holding register empty */
#define LSR_TSRE 0x20        /* transmitter shift register empty */
#define LSR_BI   0x10        /* break indicator */
#define LSR_FE   0x08        /* framing error */
#define LSR_PE   0x04        /* parity error */
#define LSR_OE   0x02        /* overrun error */
#define LSR_DR   0x01        /* data ready */
#define LSR_ERR  0x0E        /* error */

/*
 * MODEM status register
 */
#define MSR_DCD  0x80        /* data carrier detect */
#define MSR_RI   0x40        /* ring indicator */
#define MSR_DSR  0x20        /* data set ready */
#define MSR_CTS  0x10        /* clear to send */
#define MSR_DDCD 0x08        /* delta DCD */
#define MSR_TERI 0x04        /* trailing edge ring indicator */
#define MSR_DDSR 0x02        /* delta DSR */
#define MSR_DCTS 0x01        /* delta CTS */


unsigned rbr,        /* R    receive buffer register */
         thr,        /*  W   transmit holding register */
         ier,        /* RW   interrupt enable register */
         iir,        /* R    interrupt ID register */
         fcr,        /*  W   FIFO control register (16550A or later) */
         lcr,        /* RW   line control register */
         mcr,        /* RW   MODEM control register */
         lsr,        /* RW   line status register */
         msr,        /* RW   MODEM status register */
         scr,        /* RW   scratch register (16450 or later) */
         dll,        /* RW   divisor latch LSB register */
         dlm;        /* RW   divisor latch MSB register */

int
test_uart(unsigned ioadr) {
	unsigned reg_save;
	int i, done = 0, result;
	unsigned char looptest_outvals[] = {0x01, 0x02, 0x04, 0x08};
	unsigned char looptest_invals[]  = {0x20, 0x10, 0x40, 0x80};

/*
 * set up register addresses
 */
	rbr = RBR + ioadr;
	thr = THR + ioadr;
	ier = IER + ioadr;
	iir = IIR + ioadr;
	fcr = FCR + ioadr;
	lcr = LCR + ioadr;
	mcr = MCR + ioadr;
	lsr = LSR + ioadr;
	msr = MSR + ioadr;
	scr = SCR + ioadr;
	dll = DLL + ioadr;
	dlm = DLM + ioadr;

/*
 * check for no UART, indicated by loopback test failure
 */
	reg_save = inportb(mcr);
	for (i = 0; i < 4; i++) {
		/* write 1 to each MCR output bit & check corresponding MSR input bit */
		outportb(mcr, looptest_outvals[i] | MCR_LOOP);
		if ((unsigned char) (inportb(msr) & 0xF0) != looptest_invals[i]) {
			result = u_none;        /* mismatch; no UART */
			done++;
			break;
		}
	}
	outportb(mcr, reg_save);        /* restore original contents */
	if (done)
		return result;

/*
 * check for 8250, indicated by no scratch register
 */
	reg_save = inportb(scr);
	outportb(scr, 0x55);
	if ((unsigned char) inportb(scr) != 0x55) {
		result = u_8250;        /* mismatch; it must be a lowly 8250 */
		done++;
	}
	outportb(scr, 0xAA);
	if ((unsigned char) inportb(scr) != 0xAA) {
		result = u_8250;        /* mismatch; it must be a lowly 8250 */
		done++;
	}
	outportb(scr, reg_save);        /* restore original contents */
	if (done)
		return result;

/*
 * check for 16450, indicated by no "1" FIFO status bits in IIR, or
 *           16550, indicated by one "1" FIFO status bit, or
 *           16550A, indicated by two "1" FIFO status bits
 */
	reg_save = inportb(fcr);
	outportb(fcr, 1);
	switch (inportb(iir) & 0xC0) {
		case 0x00: result = u_16450;
		           break;
		case 0x80: result = u_16550;
		           break;
		case 0xC0: result = u_16550a;
		           break;
		default:   result = u_unknown;
		           break;
	}
	outportb(fcr, reg_save);        /* restore original contents */
	return result;
}


int
main(int argc, char *argv[]) {
	unsigned adr, adrs[] = {0x3F8, 0x2F8, 0x3E8, 0x2E8, 0};
	char *uarts[] = {"unknown", "none", "8250", "16450", "16650", "16650A"};
	int i, x;

	if (argc > 1) {
		for (i = 1; i < argc; i++) {
			adr = 0;
			sscanf(argv[i], "%x", &adr);
			if (adr < 0x100)
				continue;
			x = test_uart(adr);
			printf("%Xh:  %s\n", adr, uarts[x+1]);
		}
	} else {
		for (i = 0; adrs[i]; i++) {
			x = test_uart(adrs[i]);
			printf("%Xh (port %d):  %s\n", adrs[i], i + 1, uarts[x+1]);
		}
	}
	return 0;
}
